@@ -1,12 +1,58 @@
 # pylint: disable=global-variable-undefined, invalid-name
 """Multiprocess dataloader for distributed training"""
+import multiprocessing as mp
+from queue import Queue
+import traceback
+
 from .dist_context import get_sampler_pool
 from .. import backend as F
 
 __all__ = ["DistDataLoader"]
 
-DATALOADER_ID = 0
 
+def call_collate_fn(name, next_data):
+    """Call collate function"""
+    try:
+        result = DGL_GLOBAL_COLLATE_FNS[name](next_data)
+        DGL_GLOBAL_MP_QUEUES[name].put(result)
+    except Exception as e:
+        traceback.print_exc()
+        print(e)
+        raise e
+    return 1
+
+DGL_GLOBAL_COLLATE_FNS = {}
+DGL_GLOBAL_MP_QUEUES = {}
+
+def init_fn(barrier, name, collate_fn, queue):
+    """Initialize setting collate function and mp.Queue in the subprocess"""
+    global DGL_GLOBAL_COLLATE_FNS
+    global DGL_GLOBAL_MP_QUEUES
+    DGL_GLOBAL_MP_QUEUES[name] = queue
+    DGL_GLOBAL_COLLATE_FNS[name] = collate_fn
+    barrier.wait()
+    return 1
+
+def cleanup_fn(barrier, name):
+    """Clean up the data of a dataloader in the worker process"""
+    global DGL_GLOBAL_COLLATE_FNS
+    global DGL_GLOBAL_MP_QUEUES
+    del DGL_GLOBAL_MP_QUEUES[name]
+    del DGL_GLOBAL_COLLATE_FNS[name]
+    # sleep here is to ensure this function is executed in all worker processes
+    # probably need better solution in the future
+    barrier.wait()
+    return 1
+
+
+def enable_mp_debug():
+    """Print multiprocessing debug information. This is only
+    for debug usage"""
+    import logging
+    logger = mp.log_to_stderr()
+    logger.setLevel(logging.DEBUG)
+
+DATALOADER_ID = 0
 
 class DistDataLoader:
     """DGL customized multiprocessing dataloader.
@@ -66,12 +112,16 @@ class DistDataLoader:
         self.pool, self.num_workers = get_sampler_pool()
         if queue_size is None:
             queue_size = self.num_workers * 4 if self.num_workers > 0 else 4
-        self.queue_size = queue_size  # prefetch size
+        self.queue_size = queue_size
         self.batch_size = batch_size
         self.num_pending = 0
         self.collate_fn = collate_fn
         self.current_pos = 0
-        self.queue = [] # Only used when pool is None
+        if self.pool is not None:
+            m = mp.Manager()
+            self.queue = m.Queue(maxsize=queue_size)
+        else:
+            self.queue = Queue(maxsize=queue_size)
         self.drop_last = drop_last
         self.recv_idxs = 0
         self.shuffle = shuffle
@@ -89,24 +139,34 @@ class DistDataLoader:
         DATALOADER_ID += 1
 
         if self.pool is not None:
-            self.pool.set_collate_fn(self.collate_fn, self.name)
+            results = []
+            barrier = m.Barrier(self.num_workers)
+            for _ in range(self.num_workers):
+                results.append(self.pool.apply_async(
+                    init_fn, args=(barrier, self.name, self.collate_fn, self.queue)))
+            for res in results:
+                res.get()
 
     def __del__(self):
         # When the process exits, the process pool may have been closed. We should try
         # and get the process pool again and see if we need to clean up the process pool.
         self.pool, self.num_workers = get_sampler_pool()
         if self.pool is not None:
-            self.pool.delete_collate_fn(self.name)
+            results = []
+            # Here we need to create the manager and barrier again.
+            m = mp.Manager()
+            barrier = m.Barrier(self.num_workers)
+            for _ in range(self.num_workers):
+                results.append(self.pool.apply_async(cleanup_fn, args=(barrier, self.name,)))
+            for res in results:
+                res.get()
 
     def __next__(self):
-        if self.pool is None:
-            num_reqs = 1
-        else:
-            num_reqs = self.queue_size - self.num_pending
+        num_reqs = self.queue_size - self.num_pending
         for _ in range(num_reqs):
             self._request_next_batch()
         if self.recv_idxs < self.expected_idxs:
-            result = self._get_data_from_result_queue()
+            result = self.queue.get(timeout=1800)
             self.recv_idxs += 1
             self.num_pending -= 1
             return result
@@ -114,13 +174,6 @@ class DistDataLoader:
             assert self.num_pending == 0
             raise StopIteration
 
-    def _get_data_from_result_queue(self, timeout=1800):
-        if self.pool is None:
-            ret = self.queue.pop(0)
-        else:
-            ret = self.pool.get_result(self.name, timeout=timeout)
-        return ret
-
     def __iter__(self):
         if self.shuffle:
             self.dataset = F.rand_shuffle(self.dataset)
@@ -134,10 +187,10 @@ class DistDataLoader:
         if next_data is None:
             return
         elif self.pool is not None:
-            self.pool.submit_task(self.name, next_data)
+            self.pool.apply_async(call_collate_fn, args=(self.name, next_data, ))
         else:
             result = self.collate_fn(next_data)
-            self.queue.append(result)
+            self.queue.put(result)
         self.num_pending += 1
 
     def _next_data(self):
